import argparse
import sys

import numpy as np

import westpa

from westpa.core.segment import Segment
from westpa.oldtools.aframe import WESTAnalysisTool, WESTDataReaderMixin, CommonOutputMixin

import logging

log = logging.getLogger('w_succ')


class WSucc(CommonOutputMixin, WESTDataReaderMixin, WESTAnalysisTool):
    def __init__(self):
        super().__init__()
        self.include_args['CommonOutputMixin']['print_bin_labels'] = False
        self.output_file = sys.stdout

    def find_successful_trajs(self):
        pcoord_formats = {
            'u8': '%20d',
            'i8': '%20d',
            'u4': '%10d',
            'i4': '%11d',
            'u2': '%5d',
            'i2': '%6d',
            'f4': '%14.7g',
            'f8': '%23.15g',
        }

        if not self.output_suppress_headers:
            self.output_file.write(
                '''\
# successful (recycled) segments
# column 0:    iteration
# column 1:    seg_id
# column 2:    weight
# column>2:    final progress coordinate value
'''
            )
        for n_iter in range(1, self.data_manager.current_iteration):
            seg_index = self.get_seg_index(n_iter)
            all_seg_ids = np.arange(len(seg_index), dtype=np.int_)
            recycled_seg_ids = all_seg_ids[seg_index[:]['endpoint_type'] == Segment.SEG_ENDPOINT_RECYCLED]

            if len(recycled_seg_ids) == 0:
                # Attemping to retrieve a 0-length selection from HDF5 (the pcoords below) fails
                continue

            pcoord_ds = self.get_pcoord_dataset(n_iter)
            pcoord_len = pcoord_ds.shape[1]
            pcoord_ndim = pcoord_ds.shape[2]
            final_pcoords = self.get_pcoord_dataset(n_iter)[recycled_seg_ids, pcoord_len - 1, :]
            # The above HDF5 selection always returns a vector; we want a 2-d array
            final_pcoords.shape = (len(recycled_seg_ids), pcoord_ndim)

            for ipc, seg_id in enumerate(recycled_seg_ids):
                self.output_file.write('%8d    %8d    %20.14g' % (n_iter, seg_id, seg_index[seg_id]['weight']))
                fields = ['']
                for field in final_pcoords[ipc]:
                    fields.append(pcoord_formats.get(field.dtype.str[1:], '%s') % field)
                self.output_file.write('    '.join(fields))
                self.output_file.write('\n')


def entry_point():
    wsucc = WSucc()

    parser = argparse.ArgumentParser(
        'w_succ',
        description='''\
    List segments which successfully reach a target state''',
    )

    westpa.rc.add_args(parser)
    wsucc.add_args(parser)

    parser.add_argument(
        '-o',
        '--output',
        dest='output_file',
        help='Store output in OUTPUT_FILE (default: write to standard output).',
        type=argparse.FileType('wt'),
        default=sys.stdout,
    )

    args = parser.parse_args()
    westpa.rc.process_args(args, config_required=False)
    wsucc.process_args(args)
    wsucc.output_file = args.output_file

    wsucc.find_successful_trajs()


if __name__ == '__main__':
    entry_point()



import warnings

import numpy as np
import scipy.sparse as sp

from westpa.core.data_manager import weight_dtype
from westpa.core import h5io

# From postanalysis matrix
from westpa.core.binning import index_dtype
from westpa.core.reweight import stats_process


warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.filterwarnings('ignore', category=FutureWarning)


def calc_stats(bin_assignments, weights, fluxes, populations, trans, mask, sampling_frequency):
    fluxes.fill(0.0)
    populations.fill(0.0)
    trans.fill(0)

    stats_process(bin_assignments, weights, fluxes, populations, trans, mask, interval=sampling_frequency)


class FluxMatrix:
    def w_postanalysis_matrix(self):
        pi = self.progress.indicator
        pi.new_operation('Initializing')

        self.data_reader.open('r')
        nbins = self.assignments_file.attrs['nbins']

        state_labels = self.assignments_file['state_labels'][...]
        # state_map = self.assignments_file['state_map'][...]
        nstates = len(state_labels)

        start_iter, stop_iter = self.iter_range.iter_start, self.iter_range.iter_stop  # h5io.get_iter_range(self.assignments_file)
        iter_count = stop_iter - start_iter

        nfbins = nbins * nstates

        flux_shape = (iter_count, nfbins, nfbins)
        pop_shape = (iter_count, nfbins)

        h5io.stamp_iter_range(self.output_file, start_iter, stop_iter)

        bin_populations_ds = self.output_file.create_dataset('bin_populations', shape=pop_shape, dtype=weight_dtype)
        h5io.stamp_iter_range(bin_populations_ds, start_iter, stop_iter)
        h5io.label_axes(bin_populations_ds, ['iteration', 'bin'])

        flux_grp = self.output_file.create_group('iterations')
        self.output_file.attrs['nrows'] = nfbins
        self.output_file.attrs['ncols'] = nfbins

        fluxes = np.empty(flux_shape[1:], weight_dtype)
        populations = np.empty(pop_shape[1:], weight_dtype)
        trans = np.empty(flux_shape[1:], np.int64)

        # Check to make sure this isn't a data set with target states
        # tstates = self.data_reader.data_manager.get_target_states(0)
        # if len(tstates) > 0:
        #    raise ValueError('Postanalysis reweighting analysis does not support WE simulation run under recycling conditions')

        pi.new_operation('Calculating flux matrices', iter_count)
        # Calculate instantaneous statistics
        for iiter, n_iter in enumerate(range(start_iter, stop_iter)):
            # Get data from the main HDF5 file
            iter_group = self.data_reader.get_iter_group(n_iter)
            seg_index = iter_group['seg_index']
            nsegs, npts = iter_group['pcoord'].shape[0:2]
            weights = seg_index['weight']

            # Get bin and traj. ensemble assignments from the previously-generated assignments file
            assignment_iiter = h5io.get_iteration_entry(self.assignments_file, n_iter)
            bin_assignments = np.require(
                self.assignments_file['assignments'][assignment_iiter + np.s_[:nsegs, :npts]], dtype=index_dtype
            )

            mask_unknown = np.zeros_like(bin_assignments, dtype=np.uint16)

            macrostate_iiter = h5io.get_iteration_entry(self.assignments_file, n_iter)
            macrostate_assignments = np.require(
                self.assignments_file['trajlabels'][macrostate_iiter + np.s_[:nsegs, :npts]], dtype=index_dtype
            )

            # Transform bin_assignments to take macrostate membership into account
            bin_assignments = nstates * bin_assignments + macrostate_assignments

            mask_indx = np.where(macrostate_assignments == nstates)
            mask_unknown[mask_indx] = 1

            # Calculate bin-to-bin fluxes, bin populations and number of obs transitions
            calc_stats(bin_assignments, weights, fluxes, populations, trans, mask_unknown, self.sampling_frequency)

            # Store bin-based kinetics data
            bin_populations_ds[iiter] = populations

            # Setup sparse data structures for flux and obs
            fluxes_sp = sp.coo_matrix(fluxes)
            trans_sp = sp.coo_matrix(trans)

            assert fluxes_sp.nnz == trans_sp.nnz

            flux_iter_grp = flux_grp.create_group('iter_{:08d}'.format(n_iter))
            flux_iter_grp.create_dataset('flux', data=fluxes_sp.data, dtype=weight_dtype)
            flux_iter_grp.create_dataset('obs', data=trans_sp.data, dtype=np.int32)
            flux_iter_grp.create_dataset('rows', data=fluxes_sp.row, dtype=np.int32)
            flux_iter_grp.create_dataset('cols', data=fluxes_sp.col, dtype=np.int32)
            flux_iter_grp.attrs['nrows'] = nfbins
            flux_iter_grp.attrs['ncols'] = nfbins

            # Do a little manual clean-up to prevent memory explosion
            del iter_group, weights, bin_assignments
            del macrostate_assignments

            pi.progress += 1

            # Check and save the number of intermediate time points; this will be used to normalize the
            # flux and kinetics to tau in w_postanalysis_reweight.
            if self.assignments_file.attrs['subsampled'] is True or self.sampling_frequency == 'iteration':
                self.output_file.attrs['npts'] = 2
            else:
                # self.output_file.attrs['npts'] = npts if self.sampling_frequency == 'timepoint' else 2
                self.output_file.attrs['npts'] = npts



import logging

log = logging.getLogger(__name__)

import numpy


class TrajWalker:
    """A class to perform analysis by walking the trajectory tree.  A stack is used rather than recursion, or else
    the highest number of iterations capable of being considered would be the same as the Python recursion limit.
    """

    def __init__(self, data_reader, history_chunksize=100):
        self.data_reader = data_reader
        self.history_chunksize = history_chunksize
        self.n_segs_visited = 0

    # TrajTree.count_segs_in_range() is now DataReader.total_segs_in_range()

    def trace_to_root(self, n_iter, seg_id):
        '''Trace the given segment back to its starting point, returning a list of Segment
        objects describing the entire trajectory.'''

        segments = []
        segment = self.data_reader.get_segments_by_id(n_iter, [seg_id])[0]
        segments.append(segment)
        while segment.p_parent_id >= 0:
            segment = self.data_reader.get_segments_by_id(segment.n_iter - 1, [segment.p_parent_id])[0]
            segments.append(segment)
        return list(reversed(segments))

    def get_trajectory_roots(self, first_iter, last_iter, include_pcoords=True):
        '''Get segments which start new trajectories.  If min_iter or max_iter is specified, restrict the
        set of iterations within which the search is conducted.'''

        roots = []
        for n_iter in range(first_iter, last_iter + 1):
            seg_ids = self.data_reader.get_created_seg_ids(n_iter)
            segments = self.data_reader.get_segments_by_id(n_iter, seg_ids, include_pcoords=include_pcoords)
            roots.extend(segments)
        return roots

    def get_initial_nodes(self, first_iter, last_iter, include_pcoords=True):
        '''Get segments with which to begin a tree walk -- those alive or created within [first_iter,last_iter].'''

        root_ids = dict()

        # All trajectories alive or newly created in first_iter are initial nodes
        root_ids[first_iter] = set(self.data_reader.get_seg_ids(first_iter))

        # Find trajectories created in [first_iter, last_iter]
        for n_iter in range(first_iter, last_iter + 1):
            seg_ids = self.data_reader.get_created_seg_ids(n_iter)
            try:
                root_ids[n_iter].update(seg_ids)
            except KeyError:
                root_ids[n_iter] = set(seg_ids)

        # Convert to Segment objects
        segments = []
        for n_iter, id_set in root_ids.items():
            segments.extend(self.data_reader.get_segments_by_id(n_iter, id_set, include_pcoords=include_pcoords))
        return segments

    def trace_trajectories(
        self, first_iter, last_iter, callable, include_pcoords=True, cargs=None, ckwargs=None, get_state=None, set_state=None
    ):
        """
        Walk the trajectory tree depth-first, calling
          ``callable(segment, children, history, *cargs, **ckwargs)`` for each segment
        visited. ``segment`` is the segment being visited, ``children`` is that
        segment's children, ``history`` is the chain of segments leading
        to ``segment`` (not including ``segment``). get_state and set_state are
        used to record and reset, respectively, any state specific to
        ``callable`` when a new branch is traversed.
        """

        cargs = cargs or tuple()
        ckwargs = ckwargs or dict()

        # Either both or neither of external state getter/setter required
        if (get_state or set_state) and not (get_state and set_state):
            raise ValueError('either both or neither of get_state/set_state must be specified')

        # This will grow to contain the maximum trajectory length
        history = numpy.empty((self.history_chunksize,), numpy.object_)
        roots = self.get_initial_nodes(first_iter, last_iter, include_pcoords)

        for root in roots:
            children = self.data_reader.get_children(root, include_pcoords)

            # Visit the root node of each tree unconditionally
            callable(root, children, [], *cargs, **ckwargs)
            self.n_segs_visited += 1

            state_stack = [{'node': root, 'children': children, 'len_history': 0, 'ext': get_state() if get_state else None}]

            # Walk the tree, depth-first
            while state_stack:
                state = state_stack.pop(-1)

                node = state['node']
                children = state['children']
                len_history = state['len_history']
                if set_state:
                    set_state(state['ext'])

                # Descend as far as we can
                while node.n_iter < last_iter and len(children):
                    # Save current state before descending
                    state_stack.append(
                        {'node': node, 'children': children, 'len_history': len_history, 'ext': get_state() if get_state else None}
                    )

                    # Add an item to the historical record
                    if len_history >= history.shape[0]:
                        history.resize((history.shape[0] + self.history_chunksize,))
                    history[len_history] = node
                    len_history += 1

                    node = children.pop(-1)
                    children = self.data_reader.get_children(node, include_pcoords)

                    # Visit the new node as we descend
                    callable(node, children, history[:len_history], *cargs, **ckwargs)
                    self.n_segs_visited += 1



'''
Function(s) for the postanalysis toolkit
'''

import logging

from . import _reweight
from ._reweight import stats_process, reweight_for_c
from .matrix import FluxMatrix

__all__ = ['_reweight', 'stats_process', 'reweight_for_c', 'FluxMatrix']


log = logging.getLogger(__name__)



'''westext.weed -- Support for weighted ensemble equilibrium dynamics

Initial code by Dan Zuckerman (May 2011), integration by Matt Zwier,
and testing by Carsen Stringer. Re-factoring and optimization of probability
adjustment routines by Joshua L. Adelman (January 2012).
'''

from . import BinCluster, ProbAdjustEquil, UncertMath, weed_driver  # noqa
from .ProbAdjustEquil import probAdjustEquil  # noqa
from .weed_driver import WEEDDriver  # noqa
